import numpy as np
import pandas as pd
from scipy.stats import logistic
from pkg_resources import resource_filename


def load_sample_data(timevary):
    """Load data that is part of the zepid package. This data set comes from simulated data from Jessie Edwards
    (thanks Jess!). This data is used for examples on zepid.readthedocs

    Parameters
    --------------
    timevary : bool
        Whether to return the time-varying data set or the time fixed. If True then returns data set with repeated
        visits. If False then a data set with single observation per subject representing the 45-week risk is returned

    Notes
    -------------
    For the time-varying data set, the following variables are returned;
        id - participant unique ID
        enter - start of follow-up period
        out - end of time period
        male - indicator variable for male (1 = yes)
        age0 - age at enter = 0
        cd40 - CD4 T cell count at enter = 0
        dvl0 - detectable viral load data at enter = 0
        cd4 - CD4 T cell count at enter = t
        dvl - viral load at enter = t
        art - indicator of whether ART was prescribed at enter = t
        drop - indicator of whether individual dropped out of the study at enter = t (1 = yes)
        dead - indicator for death at out = t (1 = yes)

    For the time-fixed data set, the following variables are returned
        id - participant unique ID
        male - indicator variable for male (1 = yes)
        age0 - age at enter = 0
        cd40 - CD4 T cell count at enter = 0
        dvl0 - detectable viral load data at enter = 0
        art - indicator of whether ART was prescribed at enter = 0
        t - total time contributed

    Returns
    ----------
    DataFrame
        Returns either a time-varying or time-fixed pandas DataFrame

    Examples
    --------
    Load the time-fixed exposure data set
    >>>from zepid import load_sample_data
    >>>load_sample_data(timevary=False)

    Load the time-varying exposure data set
    >>>load_sample_data(timevary=True)
    """
    cols = ['id', 'enter', 'out', 'male', 'age0', 'cd40', 'dvl0', 'cd4', 'dvl', 'art', 'drop', 'dead']
    df = pd.read_csv(resource_filename('zepid', 'datasets/data.dat'),
                     delim_whitespace=True, header=None, names=cols, index_col=False)
    df.sort_values(by=['id', 'enter'], inplace=True)
    if timevary:
        return df
    else:
        dfi = df.loc[df.groupby('id').cumcount() == 0, ['id', 'male', 'age0', 'cd40', 'dvl0', 'art']].copy()
        dfo = df.loc[df.id != df.id.shift(-1)][['id', 'dead', 'drop', 'out']].copy()
        dfo.loc[(dfo['drop'] == 1) & (dfo['out'] <= 45), 'dead'] = np.nan
        dfo['dead'] = np.where((dfo['dead'] == 1) & (dfo['out'] > 45), 0, dfo['dead'])
        dff = pd.merge(dfi, dfo, left_on='id', right_on='id')
        dff.rename(columns={'out': 't'}, inplace=True)
        dff.drop('drop', axis=1, inplace=True)

        # Adding last CD4 T-cell count for continuous outcome data
        dfl = df.loc[df['out'] <= 45, ['id', 'cd4']].copy()
        dfll = dfl.loc[dfl.id != dfl.id.shift(-1)].copy()
        dfll.rename(columns={'cd4': 'cd4_wk45'}, inplace=True)
        dff = pd.merge(dff, dfll, left_on='id', right_on='id')
        dff['cd4_wk45'] = np.where(dff['dead'] == 1, np.nan, dff['cd4_wk45'])

        return dff


def load_ewing_sarcoma_data():
    """Loads Ewing's Sarcoma survival data from Glaubiger DL, Makuch R, Schwarz J, Levine AS, Johnson RE. Determination
    of prognostic factors and their influence on therapeutic results in patients with Ewing's sarcoma. Cancer.
    1980;45(8):2213-9. In total, data for 76 patients is loaded. Further descriptions of the 76 patients and the data
    can be found in Makuch RW. Adjusted survival curve estimation using covariates. J Chronic Dis. 1982;35(6):437-43.
    or Cole SR, Hernán MA. Adjusted survival curves with inverse probability weights. Comput Methods Programs
    Biomed. 2004;75(1):45-9.

    Notes
    -----------
    Variables within the dataset are
        treat - treatment (1 is novel treatment; 0 is one of three standard treatments)
        ldh - pre-treatment serum lactic acid dehydrogenase (LDH) (1 is >= 200 international units; 0 is <200)
        time - days till recurrence or censoring (continuous)
        outcome - sarcoma recurrence (1 is recurrence; 0 is censored)

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    return pd.read_csv(resource_filename('zepid', 'datasets/ewing.dat'), index_col=False)


def load_gvhd_data():
    """Loads bone marrow transplant recipient data from Keil AP, Edwards JK, Richardson DB, Naimi AI, Cole SR. The
    parametric g-formula for time-to-event data: intuition and a worked example. Epidemiology. 2014;25(6):889-97.
    Patients were followed until death or administrative censoring at 5-years.

    Notes
    ---------------
    Variables are formatted exactly as described in Keil et al. 2014
        id: unique ID for each participant
        age: participant baseline age
        agesq: squared baseline age
        agecurs1: restricted cubic spline knot 1 for baseline age
        agecurs2: restricted cubic spline knot 2 for basline age
        male: participant gender (1 is male, 0 is female)
        cmv: cytomegalovirus baseline immune status (1 is yes, 0 is no)
        all: at this time, I am unsure what this variable indicates (1, 0)
        wait: wait time from diagnosis to transplantation (months)
        day: day since transplantation
        daysq: squared day since transplantation
        daycu: cubic day since transplantation
        daycurs1: restricted cubic spline knot 1 for days since transplantation
        daycurs2: restricted cubic spline knot 2 for days since transplantation
        yesterday: previous day
        tomorrow: day after
        gvhd: indicator for Graph-versus-Host Disease (1 is yes, 0 is no)
        d: indicator of death (1 is yes, 0 is no)
        relapse: indicator for relapse (1 is yes, 0 is no)
        platnorm: indicator for normal platelet count (1 is yes, 0 is no)
        censlost: indicator for censoring due to loss-to-follow-up (1 is yes, 0 is no)
        gvhdm1: indicator for previous day diagnosis of GvHD (1 is yes, 0 is no)
        relapsem1: indicator for previous day relapse (1 is yes, 0 is no)
        platnormm1: indicator for previous day normal platelet count (1 is yes, 0 is no)
        daysnogvhd: number of consecutive days without a GvHD diagnosis
        daysnorelapse: number of consecutive days without relapse
        daysnoplatnorm: number of consecutive days without normal platelet count
        daysgvhd: number of consecutive days with GvHD
        daysrelapse: number of consecutive days after relapse
        daysplatnorm: number of consecutive days with normal platelet count

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/gvhd.dat'), delim_whitespace=True, index_col=False)

    # Coding variables to match Keil et al. exactly
    df['wait'] = df['waitdays'] / 30.5
    df['agesq'] = df['age']**2
    df['agecurs1'] = ((df['age'] > 17.0) * (df['age'] - 17.0) ** 3 -
                      ((df['age'] > 30.0) * (df['age'] - 30.0) ** 3) * (41.4 - 17.0) / (41.4 - 30.0))
    df['agecurs2'] = ((df['age'] > 25.4) * (df['age'] - 25.4) ** 3 -
                      ((df['age'] > 41.4) * (df['age'] - 41.4) ** 3) * (41.4 - 25.4) / (41.4 - 30.0))

    # Expanding data into person-time periods
    df['t_int'] = df['t'].astype(int)
    lf = pd.DataFrame(np.repeat(df.values, df['t_int'] + 1, axis=0), columns=df.columns)
    lf['yesterday'] = lf.groupby('id')['t'].cumcount()
    lf['day'] = lf['yesterday'] + 1
    lf['tomorrow'] = lf['day'] + 1
    lf['daysq'] = lf['day']**2
    lf['daycu'] = lf['day']**3
    lf['daycurs1'] = (((lf['day'] > 63) * ((lf['day'] - 63) / 63) ** 3) +
                      ((lf['day'] > 716) * ((lf['day'] - 716) / 63) ** 3) * (350.0 - 63) -
                      ((lf['day'] > 350) * ((lf['day'] - 350) / 63) ** 3) * (716 - 63) / (716 - 350))
    lf['daycurs2'] = (((lf['day'] > 168) * ((lf['day'] - 168) / 63) ** 3) +
                      ((lf['day'] > 716) * ((lf['day'] - 716) / 63) ** 3) * (350 - 168) -
                      ((lf['day'] > 350) * ((lf['day'] - 350) / 63) ** 3) * (716 - 168) / (716 - 350))
    lf['tdiff'] = lf['t'] - lf['yesterday']
    lf = lf.loc[lf['tdiff'] != 0].copy()  # Removes redundant first observation

    # Recoding variables for the expanded data
    lf['d'] = (lf['day'] >= lf['t']) * lf['d_dea']
    lf['gvhd'] = (lf['day'] > lf['t_gvhd']) * 1
    lf['relapse'] = (lf['day'] > lf['t_rel']) * 1
    lf['platnorm'] = (lf['day'] > lf['t_pla']) * 1
    lf['gvhdm1'] = (lf['yesterday'] > lf['t_gvhd']) * 1
    lf['relapsem1'] = (lf['yesterday'] > lf['t_rel']) * 1
    lf['platnormm1'] = (lf['yesterday'] > lf['t_pla']) * 1

    # Censoring time point
    lf['censeof'] = np.where((lf['day'] == lf['t']) & (lf['d'] == 0) & (lf['day'] == 1825), 1, 0)
    lf['censlost'] = np.where((lf['day'] == lf['t']) & (lf['d'] == 0) & (lf['day'] != 1825), 1, 0)

    # Setting initial values for days from X
    lf['daysnorelapse'] = np.where(lf['relapse'] == 0, lf['day'], np.nan)
    lf['daysnorelapse'] = lf['daysnorelapse'].fillna(method='ffill')
    lf['daysnoplatnorm'] = np.where(lf['id'] != lf['id'].shift(1), 0, np.nan)
    lf['daysnoplatnorm'] = np.where(lf['platnorm'] == 0, lf['day'], lf['daysnoplatnorm'])
    lf['daysnoplatnorm'] = lf['daysnoplatnorm'].fillna(method='ffill')
    lf['daysnogvhd'] = np.where(lf['gvhd'] == 0, lf['day'], np.nan)
    lf['daysnogvhd'] = lf['daysnogvhd'].fillna(method='ffill')

    lf['daysrelapse'] = lf.groupby('id')['relapse'].cumsum()
    lf['daysplatnorm'] = lf.groupby('id')['platnorm'].cumsum()
    lf['daysgvhd'] = lf.groupby('id')['gvhd'].cumsum()
    return lf[['id', 'age', 'agesq', 'agecurs1', 'agecurs2', 'male', 'cmv', 'all', 'wait', 'yesterday', 'tomorrow',
               'day', 'daysq', 'daycu', 'daycurs1', 'daycurs2', 'd', 'gvhd', 'relapse', 'platnorm', 'gvhdm1',
               'relapsem1', 'platnormm1', 'censlost', 'daysnorelapse', 'daysnoplatnorm', 'daysnogvhd',
               'daysrelapse', 'daysplatnorm', 'daysgvhd']]


def load_sciatica_data():
    """Loads the Sciatica Trial data published in; Mertens, BJA, Jacobs, WCH, Brand, R, and Peul, WC. Assessment of
    patient-specific surgery effect based on weighted estimation and propensity scoring in the re-analysis of the
    Sciatica Trial. PLOS One 2014. Details of the original Sciatica Trial are available in; Peul WC, van Houwelingen HC,
    van den Hout WB, et al. Surgery versus Prolonged Conservative Treatment for Sciatica. NEJM 2007
    DOI: 10.1177/0962280214545529

    Notes
    -------------
    Variables included are
        id: unique identifier for patient
        tpoints: follow-up time period
        time: follow-up time
        age_b: age at follow-up time
        age_t: age at randomization
        vas1_t: VAS score
        vas2_t: VAS score
        roland_t: Roland score
        likert_t: Likert score
        vas1_b: VAS1 at baseline
        vas2_b: VAS2 at baseline
        roland_b: Roland score at baseline
        likert_b: Likert score at baseline
        male: participant gender (1 is male, 0 is female)
        weight: participant weight in kilograms
        height: participant height in meters
        surgery: whether participant received the surgery (1 is surgery, 0 is not yet surgery)

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/sciatica.dat'), delim_whitespace=True, index_col=False,
                     header=None, names=['id', 'tpoints', 'time', 'age_b', 'age_t', 'vas1_t', 'vas2_t', 'roland_t',
                                         'likert_t', 'vas1_b', 'vas2_b', 'roland_b', 'likert_b', 'male', 'weight',
                                         'height', 'surgery'])
    return df


def load_leukemia_data():
    """Loads data from Freireich EJ et al., "The Effect of 6-Mercaptopurine on the Duration of Steriod-induced
    Remissions in Acute Leukemia: A Model for Evaluation of Other Potentially Useful Therapy" Blood 1963

    Notes
    ------------
    Variables included are
        t: time
        status: event indicator (0: censored, 1: relapsed)
        sex: male, female
        logwbc: log-transformed white blood cell count
        treat: treatment indicator

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/leukemia.dat'), delim_whitespace=True, index_col=False)
    return df


def load_binge_drinking_data():
    """Loads data from Ahren J et al., "Predicting the Population Health Impacts of Community Interventions: The Case
    of Alcohol Outlets and Binge Drinking" AJPH 2016. Below is some notes taken from the supplementary materials
    detailed by the paper authors;

    "The data provided for use with this sample code are simulated. The data are designed to be similar to the real
    data and associations examined in the main paper. There are 4000 observations representing individuals who are
    nested in 44 communities (variable name: neighborhood_id). The exposure of interest is neighborhood alcohol outlet
    density (alc_outlet_density), with values ranging from 39 to 168. The outcome of interest is a  binary indicator of
    binge drinking (binge_drink), and covariates include gender (male), age (age_categorical), marital status
    (married), education (education_categorical), and race/ethnicity (race_categorical). Alcohol outlet density and
    binge drinking were simulated as simple linear functions of the covariates. Thus, unlike the applied example, the
    relation of outlet density with binge drinking has a linear shape."

    Notes
    -----
    Variables included are
        male: gender (0: female, 1: male)
        age_categorical: age groups (not clearly defined as to what they refer to)
        married: marital status
        education_categorical: categories of education levels
        race_categorical: categories of race
        alc_outlet_density: density of alcohol outlets in the neighborhood (continuous)
        binge_drink: whether individual binge drinks (1: yes, 0: no)
        neighborhood_id: identifier for groups that individuals are nested in

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/binge.dat'), index_col=False)
    return df


def load_longitudinal_data():
    """Loads simulated longitudinal data. This longitudinal data is used to demonstrate the sequential regression
    time-varying g-formula, and the longitudinal targeted maximum likelihood estimator. The format of the returned
    file is a long data set

    Notes
    -----
    Variables included are
        A: treatment of interest
        Y: outcome of interest
        t: time-point
        W: baseline variable
        L: time-varying variable
        id: unique identifier for each subject

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.DataFrame()
    np.random.seed(555)
    n = 1000
    df['W'] = np.random.normal(size=n)
    df['L1'] = np.random.normal(size=n) + df['W']
    df['A1'] = np.random.binomial(1, size=n, p=logistic.cdf(df['L1']))
    df['Y1'] = np.random.binomial(1, size=n, p=logistic.cdf(-1 + 0.7 * df['L1'] - 0.3 * df['A1']))

    df['L2'] = 0.5 * df['L1'] - 0.9 * df['A1'] + np.random.normal(size=n)
    df['A2'] = np.random.binomial(1, size=n, p=logistic.cdf(1.5 * df['A1'] + 0.8 * df['L2']))
    df['A2'] = np.where(df['A1'] == 1, 1, df['A2'])
    df['Y2'] = np.random.binomial(1, size=n, p=logistic.cdf(-1 + 0.7 * df['L2'] - 0.3 * df['A2']))
    df['Y2'] = np.where(df['Y1'] == 1, np.nan, df['Y2'])

    df['L3'] = 0.5 * df['L2'] - 0.9 * df['A2'] + np.random.normal(size=n)
    df['A3'] = np.random.binomial(1, size=n, p=logistic.cdf(1.5 * df['A2'] + 0.8 * df['L3']))
    df['A2'] = np.where(df['A1'] == 1, 1, df['A2'])
    df['Y3'] = np.random.binomial(1, size=n, p=logistic.cdf(-1 + 0.7 * df['L3'] - 0.3 * df['A3']))
    df['Y3'] = np.where((df['Y2'] == 1) | (df['Y1'] == 1), np.nan, df['Y3'])

    df['id'] = df.index

    d1 = df[['id', 'Y1', 'A1', 'L1', 'W']].copy()
    d1.rename(mapper={'Y1': 'Y', 'A1': 'A', 'L1': 'L'}, axis='columns', inplace=True)
    d1['t'] = 1
    d2 = df[['id', 'Y2', 'A2', 'L2', 'W']].copy()
    d2.rename(mapper={'Y2': 'Y', 'A2': 'A', 'L2': 'L'}, axis='columns', inplace=True)
    d2['t'] = 2
    d3 = df[['id', 'Y3', 'A3', 'L3', 'W']].copy()
    d3.rename(mapper={'Y3': 'Y', 'A3': 'A', 'L3': 'L'}, axis='columns', inplace=True)
    d3['t'] = 3

    df = pd.concat([d1, d2, d3], sort=False).dropna()
    return df


def load_case_control_data():
    """Loads example case-control data, meant to reflect observed data. The goal of this case-control is to try to
    determine the causative agent / route

    Notes
    -----
    Variables included are:
        id: id number
        case: whether individual is a case or control (0: control, 1:case)
        breakfast: whether individual ate breakfast (0: no, 1: yes)
        lunch: lunch eaten by individual (0: no lunch, 1: salad, 2: chicken)
        snack: whether individual ate snack (0: no, 1:yes)
        dinner: whether individual ate dinner (0: no, 1:yes)
        sleep: hours of sleep last night (0: 6-8 hours, 1: 8+ hours)
        sunscreen: sunscreen usage (0: no, 1: yes)
        hand_wash: hand hygiene habits (1: none, 2: poor, 3: average, 4: high, 5: optimal)

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/case_control.dat'), delim_whitespace=True, index_col=False)
    return df


def load_monotone_missing_data():
    """Loads simulated data the is monotone missing. This data is for demonstration with inverse probability of missing
    weights with data that is has a monotone missing pattern.

    Notes
    -----
    Variables included are:
        id: id number
        A: variable with no missing data
        B: variable with missing data, conditional on A and L
        C: variable with missing data, conditional on B and L
        L: variable related to the missingness mechanism

    The above data is monotone missing with all individuals missing C are also missing B.

    True means in the observed data:
        B: 0.4258
        C: 0.5343

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/monotone.dat'), index_col=False)
    return df[['id', 'A', 'B', 'C', 'L']]



from .utils import (risk_ci, incidence_rate_ci, risk_ratio, risk_difference, number_needed_to_treat, odds_ratio,
                    incidence_rate_ratio, incidence_rate_difference, attributable_community_risk,
                    population_attributable_fraction, probability_to_odds, odds_to_probability, counternull_pvalue,
                    semibayes, sensitivity, specificity, ppv_converter, npv_converter, screening_cost_analyzer)



import math
import warnings
import patsy
import numpy as np
import statsmodels.api as sm
import statsmodels.formula.api as smf
from scipy.stats import logistic, norm

from zepid.causal.ipw import propensity_score
from zepid.calc import probability_to_odds


def _exposure_machine_learner(xdata, ydata, ml_model, print_results=True):
    """Function to fit machine learning predictions. Used by TMLE to generate predicted probabilities of being
    treated (i.e. Pr(A=1 | L))
    """
    # Trying to fit the Machine Learning model
    try:
        fm = ml_model.fit(X=xdata, y=ydata)
    except TypeError:
        raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                        "covers both sklearn and supylearner. If there is a predictive model you would "
                        "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                        "can work on adding support")
    if print_results and hasattr(fm, 'summarize'):  # SuPyLearner has a nice summarize function
        fm.summarize()

    # Generating predictions
    if hasattr(fm, 'predict_proba'):
        g = fm.predict_proba(xdata)[:, 1]
        return g
    elif hasattr(fm, 'predict'):
        g = fm.predict(xdata)
        return g
    else:
        raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


def _outcome_machine_learner(xdata, ydata, all_a, none_a, ml_model, continuous, print_results=True):
    """Function to fit machine learning predictions. Used by TMLE to generate predicted probabilities of outcome
    (i.e. Pr(Y=1 | A=1, L) and Pr(Y=1 | A=0, L)). Future update will include continuous Y functionality (i.e. E(Y))
    """
    # Trying to fit Machine Learning model
    try:
        fm = ml_model.fit(X=xdata, y=ydata)
    except TypeError:
        raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                        "covers both sklearn and supylearner. If there is a predictive model you would "
                        "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                        "can work on adding support")
    if print_results and hasattr(fm, 'summarize'):  # Nice summarize option from SuPyLearner
        fm.summarize()

    # Generating predictions
    if continuous:
        if hasattr(fm, 'predict'):
            qa1 = fm.predict(all_a)
            qa0 = fm.predict(none_a)
            return qa1, qa0
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")

    else:
        if hasattr(fm, 'predict_proba'):
            qa1 = fm.predict_proba(all_a)[:, 1]
            qa0 = fm.predict_proba(none_a)[:, 1]
            return qa1, qa0
        elif hasattr(fm, 'predict'):
            qa1 = fm.predict(all_a)
            qa0 = fm.predict(none_a)
            return qa1, qa0
        else:
            raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


def _missing_machine_learner(xdata, mdata, all_a, none_a, ml_model, print_results=True):
    """Function to fit machine learning predictions. Used by TMLE to generate predicted probabilities of missing
     outcome data, Pr(M=1|A,L)
    """
    # Trying to fit the Machine Learning model
    try:
        fm = ml_model.fit(X=xdata, y=mdata)
    except TypeError:
        raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                        "covers both sklearn and supylearner. If there is a predictive model you would "
                        "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                        "can work on adding support")
    if print_results and hasattr(fm, 'summarize'):  # SuPyLearner has a nice summarize function
        fm.summarize()

    # Generating predictions
    if hasattr(fm, 'predict_proba'):
        ma1 = fm.predict_proba(all_a)[:, 1]
        ma0 = fm.predict_proba(none_a)[:, 1]
        return ma1, ma0
    elif hasattr(fm, 'predict'):
        ma1 = fm.predict(all_a)
        ma0 = fm.predict(none_a)
        return ma1, ma0
    else:
        raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")


class TMLE:
    def __init__(self, df, exposure, outcome, measure=None, alpha=0.05, continuous_bound=0.0005):
        """Implementation of a single time-point target maximum likelihood estimator. It uses standard
        regression models to calculate the estimate of interest. The TMLE estimator allows users are able to directly
        input predicted outcomes from other methods (like machine learning algorithms from sklearn).

        Parameters
        ----------
        df : DataFrame
            Pandas dataframe containing the variables of interest
        exposure : str
            Column label for the exposure of interest
        outcome : str
            Column label for the outcome of interest
        measure : str, optional
            Depreciated. By default risk difference, risk ratio, and odds ratio are all calculated
        alpha : int, optional
            Alpha for confidence interval level. Default is 0.05
        continuous_bound : float, optional
            Optional argument to control the bounding feature for continuous outcomes. The bounding process may result
            in values of 0,1 which are undefined for logit(x). This parameter adds or substracts from the scenarios of
            0,1 respectively. Default value is 0.0005

        Notes
        -----
        TMLE is a double robust substitution estimator. TMLE obtains the target estimate in a single step. The
        single-step TMLE is described further by van der Laan. For further details, see the listed references.

        Continuous outcomes must be bounded between 0 and 1. TMLE does this automatically for the user. Additionally,
        the average treatment effect is estimate is back converted to the original scale of Y. When scaling Y as Y*,
        some values may take the value of 0 or 1, which breaks a logit(Y*) transformation. To avoid this issue, Y* is
        bounded by the `continuous_bound` argument. The default is 0.0005, the same as R's tmle

        The following is a general outline of the estimation process for TMLE

        1. Initial estimates for Y are predicted from a regression model. Expected values for each individual are
        generated under the scenarios of all treated vs all untreated

        .. math::

            E(Y|A, L)

        2. Predicted probabilities are generated from a regression model

        .. math::

            \Pr(A=1|L)

        3. The predicted Y is merged together with the IPW using

        .. math::

            H_a(A=a,L) = \frac{I(A=1)}{\pi_1} - \frac{I(A=0)}{\pi_0}

        for each individual. Afterwards, the predicted Y is set as an offset in the following logit model and used to
        predict values under each treatment strategy after fitted

        .. math::

            logit(E(Y|A,L)) = logit(Y_a) + \sigma * H_a

        4. The targeted Psi is estimated, representing the causal effect of all treated vs. all untreated

        Confidence intervals are constructed using influence curve theory.

        Examples
        --------
        Setting up environment
        >>>from zepid import load_sample_data, spline
        >>>from zepid.causal.doublyrobust import TMLE
        >>>df = load_sample_data(False).dropna()
        >>>df[['cd4_rs1', 'cd4_rs2']] = spline(df, 'cd40', n_knots=3, term=2, restricted=True)

        Estimating TMLE using logistic regression
        >>>tmle = TMLE(df, exposure='art', outcome='dead')
        >>># Specifying exposure/treatment model
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
        >>># Specifying outcome model
        >>>tmle.outcome_model('art + male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
        >>># TMLE estimation procedure
        >>>tmle.fit()
        >>># Printing main results
        >>>tmle.summary()
        >>># Extracting risk difference and confidence intervals, respectively
        >>>tmle.risk_difference
        >>>tmle.risk_difference_ci

        Estimating TMLE with machine learning algorithm from sklearn
        >>>from sklearn.linear_model import LogisticRegression
        >>>log1 = LogisticRegression(penalty='l1', random_state=201)
        >>>tmle = TMLE(df, 'art', 'dead')
        >>># custom_model allows specification of machine learning algorithms
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
        >>>tmle.outcome_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
        >>>tmle.fit()

        Demonstration of estimating g-model with symmetric bounds
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', bound=0.05)

        Demonstration of estimating g-model with asymmetric bounds
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', bound=[0.05, 0.9])

        References
        ----------
        Schuler, Megan S., and Sherri Rose. "Targeted maximum likelihood estimation for causal inference in
        observational studies." American journal of epidemiology 185.1 (2017): 65-73.

        Van der Laan, Mark J., and Sherri Rose. Targeted learning: causal inference for observational and experimental
        data. Springer Science & Business Media, 2011.

        Van Der Laan, Mark J., and Daniel Rubin. "Targeted maximum likelihood learning." The International Journal of
        Biostatistics 2.1 (2006).

        Gruber, S., & van der Laan, M. J. (2011). tmle: An R package for targeted maximum likelihood estimation.
        """
        if measure is not None:
            warnings.warn('As of v0.4.2, TMLE defaults to calculate all implemented measures (RD, RR, OR)', UserWarning)

        # Going through missing data (that is not the outcome)
        if df.dropna(subset=[d for d in df.columns if d != outcome]).shape[0] != df.shape[0]:
            warnings.warn("There is missing data that is not the outcome in the data set. TMLE will drop "
                          "all missing data that is not missing outcome data. TMLE will fit "
                          + str(df.dropna(subset=[d for d in df.columns if d != outcome]).shape[0]) +
                          ' of ' + str(df.shape[0]) + ' observations', UserWarning)
            self.df = df.copy().dropna(subset=[d for d in df.columns if d != outcome]).reset_index()
            print(self.df.shape[0])
        else:
            self.df = df.copy().reset_index()

        # Checking to see if missing outcome data occurs
        self._missing_indicator = '__missing_indicator__'
        if self.df.dropna(subset=[outcome]).shape[0] != self.df.shape[0]:
            self._miss_flag = True
            self.df[self._missing_indicator] = np.where(self.df[outcome].isna(), 0, 1)
        else:
            self._miss_flag = False
            self.df[self._missing_indicator] = 1

        # Detailed steps follow "Targeted Learning" chapter 4, figure 4.2 by van der Laan, Rose
        self._exposure = exposure
        self._outcome = outcome

        if df[outcome].dropna().value_counts().index.isin([0, 1]).all():
            self._continuous_outcome = False
        else:
            self._continuous_outcome = True
            self._continuous_min = np.min(df[outcome])
            self._continuous_max = np.max(df[outcome])
            self._cb = continuous_bound
            self.df[outcome] = self._unit_bounds(y=df[outcome], mini=self._continuous_min,
                                                 maxi=self._continuous_max, bound=self._cb)

        self._out_model = None
        self._exp_model = None
        self._miss_model = None
        self._fit_exposure_model = False
        self._fit_outcome_model = False
        self._fit_missing_model = False
        self.alpha = alpha

        self.QA0W = None
        self.QA1W = None
        self.QAW = None
        self.g1W = None
        self.g0W = None
        self.m1W = None
        self.m0W = None
        self._epsilon = None

        self.risk_difference = None
        self.risk_difference_ci = None
        self.risk_ratio = None
        self.risk_ratio_ci = None
        self.odds_ratio = None
        self.odds_ratio_ci = None
        self.average_treatment_effect = None
        self.average_treatment_effect_ic = None

    def exposure_model(self, model, custom_model=None, bound=False, print_results=True):
        """Estimation of Pr(A=1|L), which is termed as g(A=1|L) in the literature

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._exp_model = self._exposure + ' ~ ' + model

        # Step 3) Estimation of g-model (exposure model)
        if custom_model is None:
            fitmodel = propensity_score(self.df, self._exp_model, print_results=print_results)
            self.g1W = fitmodel.predict(self.df)

        # User-specified prediction model
        else:
            data = patsy.dmatrix(model + ' - 1', self.df)
            self.g1W = _exposure_machine_learner(xdata=np.asarray(data), ydata=np.asarray(self.df[self._exposure]),
                                                 ml_model=custom_model, print_results=print_results)

        self.g0W = 1 - self.g1W
        if bound:  # Bounding predicted probabilities if requested
            self.g1W = self._bounding(self.g1W, bounds=bound)
        if bound:  # Bounding predicted probabilities if requested
            self.g0W = self._bounding(self.g0W, bounds=bound)

        self._fit_exposure_model = True

    def missing_model(self, model, custom_model=None, print_results=True):
        """Estimation of Pr(M=1|A,L), which is the missing data mechanism for the outcome. The corresponding observation
        probabilities are used to update the clever covariates for estimation of Qn.

        The initial estimate of Q is still based on complete observations only

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'. The treatment must be
            included for the missing data model
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        # Error if no missing outcome data
        if not self._miss_flag:
            raise ValueError("No missing outcome data is present in the data set")

        # Warning if exposure is not included in the missingness of outcome model
        if self._exposure not in model:
            warnings.warn("For the specified missing outcome model, the exposure variable should be included in the "
                          "model", UserWarning)

        self._miss_model = self._missing_indicator + ' ~ ' + model

        # Step 3b) Prediction for M if missing outcome data exists
        if custom_model is None:  # Logistic Regression model for predictions
            fitmodel = propensity_score(self.df, self._miss_model, print_results=print_results)
            dfx = self.df.copy()
            dfx[self._exposure] = 1
            self.m1W = fitmodel.predict(dfx)
            dfx = self.df.copy()
            dfx[self._exposure] = 0
            self.m0W = fitmodel.predict(dfx)

        # User-specified model
        else:
            data = patsy.dmatrix(model + ' - 1', self.df)

            dfx = self.df.copy()
            dfx[self._exposure] = 1
            adata = patsy.dmatrix(model + ' - 1', dfx)
            dfx = self.df.copy()
            dfx[self._exposure] = 0
            ndata = patsy.dmatrix(model + ' - 1', dfx)

            self.m1W, self.m0W = _missing_machine_learner(xdata=np.asarray(data),
                                                          mdata=np.asarray(self.df[self._missing_indicator]),
                                                          all_a=adata, none_a=ndata,
                                                          ml_model=custom_model, print_results=print_results)

        self._fit_missing_model = True

    def outcome_model(self, model, custom_model=None, print_results=True, continuous_distribution='gaussian'):
        """Estimation of E(Y|A,L,M=1), which is also written sometimes as Q(A,W,M=1) or Pr(Y=1|A,W,M=1). Estimation
        of this model is based on complete observations of Y only

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        continuous_distribution : str, optional
            Distribution to use for continuous outcomes. Options are 'gaussian' for normal distributions and 'poisson'
            for Poisson distributions
        """
        self._out_model = self._outcome + ' ~ ' + model

        if self._miss_flag:
            cc = self.df.copy().dropna()
        else:
            cc = self.df.copy()

        # Step 1) Prediction for Q (estimation of Q-model)
        if custom_model is None:  # Logistic Regression model for predictions
            if self._continuous_outcome:
                if continuous_distribution == 'gaussian':
                    f = sm.families.family.Gaussian()
                elif continuous_distribution == 'poisson':
                    f = sm.families.family.Poisson()
                else:
                    raise ValueError("Only 'gaussian' and 'poisson' distributions are supported")
                log = smf.glm(self._out_model, cc, family=f).fit()
            else:
                f = sm.families.family.Binomial()
                log = smf.glm(self._out_model, cc, family=f).fit()

            if print_results:
                print('\n----------------------------------------------------------------')
                print('MODEL: ' + self._out_model)
                print('-----------------------------------------------------------------')
                print(log.summary())

            # Step 2) Estimation under the scenarios
            dfx = self.df.copy()
            dfx[self._exposure] = 1
            self.QA1W = log.predict(dfx)
            dfx = self.df.copy()
            dfx[self._exposure] = 0
            self.QA0W = log.predict(dfx)

        # User-specified model
        else:
            data = patsy.dmatrix(model + ' - 1', self.df)

            dfx = self.df.copy()
            dfx[self._exposure] = 1
            adata = patsy.dmatrix(model + ' - 1', dfx)
            dfx = self.df.copy()
            dfx[self._exposure] = 0
            ndata = patsy.dmatrix(model + ' - 1', dfx)

            self.QA1W, self.QA0W = _outcome_machine_learner(xdata=np.asarray(data),
                                                            ydata=np.asarray(self.df[self._outcome]),
                                                            all_a=adata, none_a=ndata,
                                                            ml_model=custom_model,
                                                            continuous=self._continuous_outcome,
                                                            print_results=True)

        self.QAW = self.QA1W * self.df[self._exposure] + self.QA0W * (1 - self.df[self._exposure])
        self._fit_outcome_model = True

    def fit(self):
        """Estimates risk difference, risk ratio, and odds ratio based on the gAW and QAW. Confidence intervals come
        from the influence curve

        Returns
        -------
        TMLE gains Psi and confint attributes
        """
        if (self._fit_exposure_model is False) or (self._fit_outcome_model is False):
            raise ValueError('The exposure and outcome models must be specified before the psi estimate can '
                             'be generated')
        if self._miss_flag and not self._fit_missing_model:
            warnings.warn("No missing data model has been specified. All missing outcome data is assumed to be "
                          "missing completely at random. To relax this assumption to outcome data is missing at random"
                          "please use the `missing_model()` function", UserWarning)

        # Step 4) Calculating clever covariate (HAW)
        if self._miss_flag and self._fit_missing_model:
            self.g1W_total = self.g1W * self.m1W
            self.g0W_total = self.g0W * self.m0W
        else:
            self.g1W_total = self.g1W
            self.g0W_total = self.g0W
        H1W = self.df[self._exposure] / self.g1W_total
        H0W = -(1 - self.df[self._exposure]) / self.g0W_total
        HAW = H1W + H0W

        # Step 5) Estimating TMLE
        f = sm.families.family.Binomial()
        y = self.df[self._outcome]
        log = sm.GLM(y, np.column_stack((H1W, H0W)), offset=np.log(probability_to_odds(self.QAW)),
                     family=f, missing='drop').fit()
        self._epsilon = log.params
        Qstar1 = logistic.cdf(np.log(probability_to_odds(self.QA1W)) + self._epsilon[0] / self.g1W_total)
        Qstar0 = logistic.cdf(np.log(probability_to_odds(self.QA0W)) - self._epsilon[1] / self.g0W_total)
        Qstar = log.predict(np.column_stack((H1W, H0W)), offset=np.log(probability_to_odds(self.QAW)))

        # Step 6) Calculating Psi
        if self.alpha == 0.05:  # Without this, won't match R exactly. R relies on 1.96, while I use SciPy
            zalpha = 1.96
        else:
            zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        # p-values are not implemented (doing my part to enforce CL over p-values)
        delta = np.where(self.df[self._missing_indicator] == 1, 1, 0)
        if self._continuous_outcome:
            # Calculating Average Treatment Effect
            Qstar = self._unit_unbound(Qstar, mini=self._continuous_min, maxi=self._continuous_max)
            Qstar1 = self._unit_unbound(Qstar1, mini=self._continuous_min, maxi=self._continuous_max)
            Qstar0 = self._unit_unbound(Qstar0, mini=self._continuous_min, maxi=self._continuous_max)

            self.average_treatment_effect = np.nanmean(Qstar1 - Qstar0)
            # Influence Curve for CL
            y_unbound = self._unit_unbound(self.df[self._outcome], mini=self._continuous_min, maxi=self._continuous_max)
            ic = np.where(delta == 1,
                          HAW * (y_unbound - Qstar) + (Qstar1 - Qstar0) - self.average_treatment_effect,
                          Qstar1 - Qstar0 - self.average_treatment_effect)
            varIC = np.nanvar(ic, ddof=1) / self.df.shape[0]
            self.average_treatment_effect_ic = [self.average_treatment_effect - zalpha * math.sqrt(varIC),
                                                self.average_treatment_effect + zalpha * math.sqrt(varIC)]
        else:
            # Calculating Risk Difference
            self.risk_difference = np.nanmean(Qstar1 - Qstar0)
            # Influence Curve for CL
            ic = np.where(delta == 1,
                          HAW * (self.df[self._outcome] - Qstar) + (Qstar1 - Qstar0) - self.risk_difference,
                          (Qstar1 - Qstar0) - self.risk_difference)
            varIC = np.nanvar(ic, ddof=1) / self.df.shape[0]
            self.risk_difference_ci = [self.risk_difference - zalpha * np.sqrt(varIC),
                                       self.risk_difference + zalpha * np.sqrt(varIC)]

            # Calculating Risk Ratio
            self.risk_ratio = np.nanmean(Qstar1) / np.nanmean(Qstar0)
            # Influence Curve for CL
            ic = np.where(delta == 1,
                          (1/np.mean(Qstar1) * (H1W * (self.df[self._outcome] - Qstar) + Qstar1 - np.mean(Qstar1)) -
                           (1/np.mean(Qstar0)) * (-1*H0W*(self.df[self._outcome] - Qstar) + Qstar0 - np.mean(Qstar0))),
                          (Qstar1 - np.mean(Qstar1)) + Qstar0 - np.mean(Qstar0))

            varIC = np.nanvar(ic, ddof=1) / self.df.shape[0]
            self.risk_ratio_ci = [np.exp(np.log(self.risk_ratio) - zalpha * np.sqrt(varIC)),
                                  np.exp(np.log(self.risk_ratio) + zalpha * np.sqrt(varIC))]

            # Calculating Odds Ratio
            self.odds_ratio = (np.nanmean(Qstar1) / (1 - np.nanmean(Qstar1)
                                                     )) / (np.nanmean(Qstar0) / (1 - np.nanmean(Qstar0)))
            # Influence Curve for CL
            ic = np.where(delta == 1,
                          ((1/(np.nanmean(Qstar1)*(1 - np.nanmean(Qstar1))) *
                            (H1W*(self.df[self._outcome] - Qstar) + Qstar1)) -
                           (1/(np.nanmean(Qstar0)*(1 - np.nanmean(Qstar0))) *
                            (-1*H0W*(self.df[self._outcome] - Qstar) + Qstar0))),

                          ((1 / (np.nanmean(Qstar1) * (1 - np.nanmean(Qstar1))) * Qstar1 -
                           (1 / (np.nanmean(Qstar0) * (1 - np.nanmean(Qstar0))) * Qstar0))))
            varIC = np.nanvar(ic, ddof=1) / self.df.shape[0]
            self.odds_ratio_ci = [np.exp(np.log(self.odds_ratio) - zalpha * np.sqrt(varIC)),
                                  np.exp(np.log(self.odds_ratio) + zalpha * np.sqrt(varIC))]

    def summary(self, decimal=3):
        """Prints summary of model results

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if (self._fit_exposure_model is False) or (self._fit_exposure_model is False):
            raise ValueError('The exposure and outcome models must be specified before the psi estimate can '
                             'be generated')

        if self._continuous_outcome:
            print('----------------------------------------------------------------------')
            print('Average Treatment Effect: ', round(float(self.average_treatment_effect), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.average_treatment_effect_ic[0], decimal)), ',',
                  str(round(self.average_treatment_effect_ic[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
        else:
            print('----------------------------------------------------------------------')
            print('Risk Difference: ', round(float(self.risk_difference), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_difference_ci[0], decimal)), ',',
                  str(round(self.risk_difference_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Risk Ratio: ', round(float(self.risk_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_ratio_ci[0], decimal)), ',',
                  str(round(self.risk_ratio_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Odds Ratio: ', round(float(self.odds_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.odds_ratio_ci[0], decimal)), ',',
                  str(round(self.odds_ratio_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')

    @staticmethod
    def _bounding(v, bounds):
        """Background function to perform bounding feature. Not intended for users to access

        v:
            -Values to be bounded
        bounds:
            -Percentile thresholds for bounds
        """
        if type(bounds) is float:  # Symmetric bounding
            if bounds < 0 or bounds > 1:
                raise ValueError('Bound value must be between (0, 1)')
            v = np.where(v < bounds, bounds, v)
            v = np.where(v > 1-bounds, 1-bounds, v)
        elif type(bounds) is str:  # Catching string inputs
            raise ValueError('Bounds must either be a float between (0, 1), or a collection of floats between (0, 1)')
        else:  # Asymmetric bounds
            if bounds[0] > bounds[1]:
                raise ValueError('Bound thresholds must be listed in ascending order')
            if len(bounds) > 2:
                warnings.warn('It looks like your specified bounds is more than two floats. Only the first two '
                              'specified bounds are used by the bound statement. So only ' +
                              str(bounds[0:2]) + ' will be used', UserWarning)
            if type(bounds[0]) is str or type(bounds[1]) is str:
                raise ValueError('Bounds must be floats between (0, 1)')
            if (bounds[0] < 0 or bounds[1] > 1) or (bounds[0] < 0 or bounds[1] > 1):
                raise ValueError('Both bound values must be between (0, 1)')
            v = np.where(v < bounds[0], bounds[0], v)
            v = np.where(v > bounds[1], bounds[1], v)
        return v

    @staticmethod
    def _unit_bounds(y, mini, maxi, bound):
        v = (y - mini) / (maxi - mini)
        v = np.where(v < bound, bound, v)
        v = np.where(v > 1-bound, 1-bound, v)
        return v

    @staticmethod
    def _unit_unbound(ystar, mini, maxi):
        return ystar*(maxi - mini) + mini



'''Tools for sensitivity analyses. I still need to read Lash & Fox to integrate more tools
for multiple bias analysis. This branch is still very much a work in progress. The goal is 
to simplify sensitivity analyses, in the hopes they become more common in publications

-MonteCarloRR(): generates a corrected RR distribution based on binary confounder
-trapezoidal(): generates a trapezoidal distribution of values
'''


from .Simple import MonteCarloRR
from .distributions import trapezoidal



import warnings
import math
import patsy
import numpy as np
import pandas as pd
from scipy.stats.kde import gaussian_kde
from statsmodels.stats.weightstats import DescrStatsW
import matplotlib.pyplot as plt
from .utils import propensity_score

from zepid.calc import probability_to_odds


class IPTW:
    def __init__(self, df, treatment, stabilized=True, standardize='population'):
        """
        Calculates the weight for inverse probability of treatment weights through logistic regression.
        Both stabilized or unstabilized weights are implemented. Default is just to calculate the prevalence
        of the treatment in the population.

        The formula for stabilized weights is

        .. math::

            \pi_i = \frac{\Pr(A=a)}{\Pr(A=a|L=l)}

        For unstabilized weights

        .. math::

            \pi_i = \frac{1}{\Pr(A=a|L=l)}

        SMR unstabilized weights for weighting to exposed (A=1)

        .. math::

            \pi_i &= 1 if A = 1 \\
                  &= \frac{\Pr(A=1|L=l)}{\Pr(A=0|L=l)} if A = 0

        For SMR weighted to the unexposed (A=0) the equation becomes

        .. math::

            \pi_i &= \frac{\Pr(A=0|L=l)}{\Pr(A=1|L=l)} if A=1 \\
                  &= 1 if A = 0

        Parameters
        ----------
        df : DataFrame
            Pandas dataframe object containing all variables of interest
        treatment : str
            Variable name of treatment variable of interest. Must be coded as binary. 1 should indicate treatment,
            while 0 indicates no treatment
        stabilized : bool, optional
            Whether to return stabilized or unstabilized weights. Default is stabilized weights (True)
        standardize : str, optional
            Who to standardize the estimate to. Options are the entire population, the exposed, or the unexposed. See
            Sato & Matsuyama Epidemiology (2003) for details on weighting to exposed/unexposed. Weighting to the
            exposed or unexposed is also referred to as SMR weighting. Options for standardization are:
            * 'population'    :   weight to entire population
            * 'exposed'       :   weight to exposed individuals
            * 'unexposed'     :   weight to unexposed individuals

        Examples
        --------
        Stabilized IPTW weights
        >>>import zepid as ze
        >>>from zepid.causal.ipw import IPTW
        >>>df = ze.load_sample_data(False)
        >>>ipt = IPTW(df, treatment='art', stabilized=True)
        >>>ipt.regression_models('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
        >>>ipt.fit()

        Unstabilized IPTW weights
        >>>ipt = IPTW(df, treatment='art', stabilized=False)
        >>>ipt.regression_models('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
        >>>ipt.fit()

        SMR weight to the exposed population
        >>>ipt = IPTW(df, treatment='art', stabilized=False, standardize='exposed')
        >>>ipt.regression_models('male + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
        >>>ipt.fit()

        Diagnostics:
        >>>ipt.positivity()

        >>>print(ipt.standardized_mean_differences())

        >>>import matplotlib.pyplot as plt
        >>>ipt.plot_boxplot()
        >>>plt.show()

        >>>ipt.plot_kde()
        >>>plt.show()

        >>>ipt.plot_love()
        >>>plt.show()
        """
        self.denominator_model = None
        self.numerator_model = None
        self.__mdenom = None

        self.Weight = None
        self.ProbabilityNumerator = None
        self.ProbabilityDenominator = None

        self.df = df.copy()
        self.ex = treatment
        self.stabilized = stabilized
        if standardize in ['population', 'exposed', 'unexposed']:
            self.standardize = standardize
        else:
            raise ValueError('Please specify one of the currently supported weighting schemes: ' +
                             'population, exposed, unexposed')

        self._pos_avg = None
        self._pos_min = None
        self._pos_max = None
        self._pos_sd = None

    def regression_models(self, model_denominator, model_numerator='1', print_results=True,
                          custom_model_denominator=None, custom_model_numerator=None):
        """Logistic regression model(s) for propensity score models. The model denominator must be specified for both
        stabilized and unstabilized weights. The optional argument 'model_numerator' allows specification of the
        stabilization factor for the weight numerator. By default model results are returned

        Parameters
        ------------
        model_denominator : str
            String listing variables to predict the exposure, separated by +. For example, 'var1 + var2 + var3'. This
            is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the exposure, separated by +. Only used to calculate the
            numerator. Default ('1') calculates the overall probability of exposure. In general this is recommended. If
            confounding variables are included in the numerator, they would later need to be adjusted for. Argument is
            also only used when calculating stabilized weights
        print_results : bool, optional
            Whether to print the model results from the regression models. Default is True
        custom_model_denominator : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        custom_model_numerator : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities

        Notes
        -----
        If custom models are used, it is important that GEE is used to obtain the variance. Bootstrapped confidence
        intervals are incorrect with the usage of some machine learning models
        """
        # Calculating denominator probabilities
        self.__mdenom = model_denominator
        if custom_model_denominator is None:
            self.denominator_model = propensity_score(self.df, self.ex + ' ~ ' + model_denominator,
                                                      print_results=print_results)
            d = self.denominator_model.predict(self.df)
        else:
            data = patsy.dmatrix(model_denominator + ' - 1', self.df)
            try:
                fm = custom_model_denominator.fit(X=data, y=self.df[self.ex])
            except TypeError:
                raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                                "covers both sklearn and supylearner. If there is a predictive model you would "
                                "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                                "can work on adding support")
            if print_results and hasattr(fm, 'summarize'):
                fm.summarize()
            if hasattr(fm, 'predict_proba'):
                d = fm.predict_proba(data)[:, 1]
            elif hasattr(fm, 'predict'):
                d = fm.predict(data)
            else:
                raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")
            self.denominator_model = fm

        self.df['__denom__'] = d

        # Calculating numerator probabilities (if stabilized)
        if self.stabilized is True:
            if custom_model_numerator is None:
                self.numerator_model = propensity_score(self.df, self.ex + ' ~ ' + model_numerator,
                                                        print_results=print_results)
                n = self.numerator_model.predict(self.df)

            else:
                data = patsy.dmatrix(model_numerator + ' - 1', self.df)
                try:
                    fm = custom_model_numerator.fit(X=data, y=self.df[self.ex])
                except TypeError:
                    raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                                    "covers both sklearn and supylearner. If there is a predictive model you would "
                                    "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                                    "can work on adding support")
                if print_results and hasattr(fm, 'summarize'):
                    fm.summarize()
                if hasattr(fm, 'predict_proba'):
                    n = fm.predict_proba(data)[:, 1]
                elif hasattr(fm, 'predict'):
                    n = fm.predict(data)
                else:
                    raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")

        # If unstabilized, numerator is always 1
        else:
            if model_numerator != '1':
                raise ValueError('Argument for model_numerator is only used for stabilized=True')
            n = 1
        self.df['__numer__'] = n

    def fit(self):
        """Uses the specified regression models from 'regression_models' to generate the corresponding inverse
        probability of treatment weights

        Returns
        ------------
        IPTW class gains the Weight, ProbabilityDenominator, and ProbabilityNumerator attributed. Weights is a pandas
        Series containing the calculated IPTW.
        """
        if self.denominator_model is None:
            raise ValueError('No model has been fit to generated predicted probabilities')

        self.Weight = self._weight_calculator(self.df, denominator='__denom__', numerator='__numer__')
        self.ProbabilityDenominator = self.df['__denom__']
        self.ProbabilityNumerator = self.df['__numer__']
        self.df['iptw'] = self.df['__numer__'] / self.df['__denom__']

    def plot_kde(self, measure='probability', bw_method='scott', fill=True, color_e='b', color_u='r'):
        """Generates a density plot that can be used to check whether positivity may be violated qualitatively. The
        kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or Silverman's Rule can be implemented.
        Alternative option to the boxplot of probabilities

        Parameters
        ------------
        measure : str, optional
            Measure to plot. Options include either the probabilities or log-odds stratified by treatment received.
            Default is probabilities (measure='probability'). Log-odds can be requested via measure='logit'
        bw_method : str, optional
            Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
        fill : bool, optional
            Whether to color the area under the density curves. Default is true
        color_e : str, optional
            Color of the line/area for the treated group. Default is Blue
        color_u : str, optional
            Color of the line/area for the treated group. Default is Red

        Returns
        ---------------
        matplotlib axes
        """
        if measure == 'probability':
            x = np.linspace(0, 1, 10000)
            density_t = gaussian_kde(self.df.loc[self.df[self.ex] == 1]['__denom__'].dropna(),
                                     bw_method=bw_method)
            density_u = gaussian_kde(self.df.loc[self.df[self.ex] == 0]['__denom__'].dropna(),
                                     bw_method=bw_method)
        elif measure == 'logit':
            t = np.log(probability_to_odds(self.df.loc[self.df[self.ex] == 1]['__denom__'].dropna()))
            density_t = gaussian_kde(t, bw_method=bw_method)

            u = np.log(probability_to_odds(self.df.loc[self.df[self.ex] == 0]['__denom__'].dropna()))
            density_u = gaussian_kde(u, bw_method=bw_method)
            x = np.linspace(np.min((np.min(t), np.min(u))) - 1, np.max((np.max(t), np.max(u))) + 1, 10000)
        else:
            raise ValueError("Only plots of probabilities or log-odds are supported. Please specify either "
                             "'probability' or 'logit")

        ax = plt.gca()
        if fill:
            ax.fill_between(x, density_t(x), color=color_e, alpha=0.2, label=None)
            ax.fill_between(x, density_u(x), color=color_u, alpha=0.2, label=None)
        ax.plot(x, density_t(x), color=color_e, label='Treat = 1')
        ax.plot(x, density_u(x), color=color_u, label='Treat = 0')
        if measure == 'probability':
            ax.set_xlabel('Probability')
        else:
            ax.set_xlabel('Log-Odds')
        ax.set_ylabel('Density')
        ax.legend()
        return ax

    def plot_boxplot(self, measure='probability'):
        """Generates a stratified boxplot that can be used to visually check whether positivity may be violated,
        qualitatively. Alternative option to the kernel density plot.

        Parameters
        ----------
        measure : str, optional
            Measure to plot. Options include either the probabilities or log-odds stratified by treatment received.
            Default is probabilities (measure='probability'). Log-odds can be requested via measure='logit'

        Returns
        -------------
        matplotlib axes
        """
        if measure == 'probability':
            boxes = (self.df.loc[self.df[self.ex] == 1]['__denom__'].dropna(),
                     self.df.loc[self.df[self.ex] == 0]['__denom__'].dropna())

        elif measure == 'logit':
            boxes = (np.log(probability_to_odds(self.df.loc[self.df[self.ex] == 1]['__denom__'].dropna())),
                     np.log(probability_to_odds(self.df.loc[self.df[self.ex] == 0]['__denom__'].dropna())))
        else:
            raise ValueError("Only plots of probabilities or log-odds are supported. Please specify either "
                             "'probability' or 'logit")

        labs = ['Treat = 1', 'Treat = 0']
        meanpointprops = dict(marker='D', markeredgecolor='black', markerfacecolor='black')
        ax = plt.gca()
        ax.boxplot(boxes, labels=labs, meanprops=meanpointprops, showmeans=True)
        if measure == 'probability':
            ax.set_ylabel('Probability')
        else:
            ax.set_ylabel('Log-Odds')
        return ax

    def positivity(self, decimal=3):
        """Use this to assess whether positivity is a valid assumption. Note that this should only be used for
        stabilized weights generated from IPTW. This diagnostic method is based on recommendations from
        Cole SR & Hernan MA (2008). For more information, see the following paper:
        Cole SR, Hernan MA. Constructing inverse probability weights for marginal structural models.
        American Journal of Epidemiology 2008; 168(6):656–664.

        Parameters
        --------------
        decimal : int, optional
            Number of decimal places to display. Default is three

        Returns
        --------------
        None
            Prints the positivity results to the console but does not return any objects
        """
        self.df['iptw'] = self.Weight
        if not self.stabilized:
            warnings.warn('Positivity should only be used for stabilized IPTW', UserWarning)
        self._pos_avg = float(np.mean(self.df['iptw'].dropna()))
        self._pos_max = np.max(self.df['iptw'].dropna())
        self._pos_min = np.min(self.df['iptw'].dropna())
        self._pos_sd = float(np.std(self.df['iptw'].dropna()))
        print('----------------------------------------------------------------------')
        print('IPW Diagnostic for positivity')
        print('''If the mean of the weights is far from either the min or max, this may\n indicate the model is
                incorrect or positivity is violated''')
        print('Standard deviation can help in IPTW model selection')
        print('----------------------------------------------------------------------')
        print('Mean weight:\t\t\t', round(self._pos_avg, decimal))
        print('Standard Deviation:\t\t', round(self._pos_sd, decimal))
        print('Minimum weight:\t\t\t', round(self._pos_min, decimal))
        print('Maximum weight:\t\t\t', round(self._pos_max, decimal))
        print('----------------------------------------------------------------------')

    def plot_love(self, color_unweighted='r', color_weighted='b', shape_unweighted='o', shape_weighted='o'):
        """Generates a Love-plot to detail covariate balance based on the IPTW weights. Further details on the usage of
        this plot are available in Austin PC & Stuart EA 2015 https://onlinelibrary.wiley.com/doi/full/10.1002/sim.6607

        The Love plot generates a dashed line at standardized mean difference of 0.10. In general, it is recommended
        that weighted SMD are below this level. Variables above this level may be unbalanced despite the weighting
        procedure. Different functional forms (or approaches like machine learning) can be considered

        Returns
        -------
        axes
            Matplotlib axes of the Love plot
        """
        to_plot = self.standardized_mean_differences()
        to_plot['smd_w'] = np.absolute(to_plot['smd_w'])
        to_plot['smd_u'] = np.absolute(to_plot['smd_u'])
        to_plot = to_plot.sort_values(by='smd_u', ascending=True).reset_index(drop=True)

        # Generate plot
        ax = plt.gca()
        ax.plot(to_plot.smd_u, to_plot.index, shape_unweighted, c=color_unweighted, label='Unweighted')
        ax.plot(to_plot.smd_w, to_plot.index, shape_weighted, c=color_weighted, label='Weighted')
        ax.set_xlim([0, np.max([np.max(to_plot['smd_w']), np.max(to_plot['smd_u'])]) + 0.5])
        ax.set_xlabel('Absolute Standardized Difference')
        ax.axvline(0.1, color='gray')
        ax.set_yticks([i for i in range(to_plot.shape[0])])
        ax.set_yticklabels(to_plot['labels'])
        ax.legend()
        return ax

    def standardized_mean_differences(self):
        """Calculates the standardized mean differences for all variables. Default calculates the standardized mean
        difference for all variables include in the IPTW denominator

        Returns
        -------
        DataFrame
            Returns pandas DataFrame of calculated standardized mean differences. Columns are labels (variables labels),
            smd_u (unweighted standardized difference), and smd_w (weighted standardized difference)
        """
        vars = patsy.dmatrix(self.__mdenom + ' - 1', self.df, return_type='dataframe')
        w_diff = []
        u_diff = []
        vlabel = []

        # Pull out list of terms and the corresponding dataframe slice(s)
        term_dict = vars.design_info.term_name_slices

        # Looping through the terms
        for term in vars.design_info.terms:
            # Adding term labels
            vlabel.append(term.name())

            # Pulling out data corresponding to term
            chunk = term_dict[term.name()]
            v = vars.iloc[:, chunk].copy()

            # Detecting variable type
            if v.shape[1] != 1:
                vtype = 'categorical'
            elif v.dropna().isin([0, 1]).all(axis=None):
                vtype = 'binary'
            else:
                vtype = 'continuous'

            # calculate the absolute standardized difference
            dat = pd.concat([v, self.df[[self.ex, 'iptw']]], axis=1)
            wsmd = self._standardized_difference(variable=dat, var_type=vtype, weighted=True)
            w_diff.append(wsmd)
            usmd = self._standardized_difference(variable=dat, var_type=vtype, weighted=False)
            u_diff.append(usmd)

        # Setting up DataFrame to return with calculated differences
        s = pd.DataFrame()
        s['labels'] = vlabel
        s['smd_w'] = w_diff
        s['smd_u'] = u_diff
        return s

    def _standardized_difference(self, variable, var_type, weighted=True):
        """Calculates the standardized mean difference between the treat/exposed and untreated/unexposed for a
        specified variable. Useful for checking whether a confounder was balanced between the two treatment groups
        by the specified IPTW model SMD based on: Austin PC 2011; https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3144483/

        For efficiency, it is recommended you use standardized_mean_differences(). That function calculates the
        standardized mean differences for all variables included in the denominator

        Parameters
        ---------------
        variable : str, list
            Label for variable to calculate the standardized difference. If categorical variables, it should be a list
            of variable labels
        var_type : str
            Variable type. Options are 'binary' 'continuous' or 'categorical'. For categorical variable should be a
            list of columns labels
        weighted : bool, optional
            Whether to return the weighted standardized mean difference or the unweighted. Default is to return the
            weighted.

        Returns
        --------------
        None
            Prints the positivity results to the console but does not return any objects
        """
        # Pulling out relevant data
        dft = variable.loc[(variable[self.ex] == 1) & (variable['iptw'].notnull())].copy()
        dfn = variable.loc[(variable[self.ex] == 0) & (variable['iptw'].notnull())].copy()
        # removing self.ex and 'iptw' from vars to calculate for
        vcols = list(variable.columns)
        vcols.remove(self.ex)
        vcols.remove('iptw')

        if var_type == 'binary':
            if weighted:
                dwt = DescrStatsW(dft[vcols], weights=dft['iptw'])
                wt = dwt.mean
                dwn = DescrStatsW(dfn[vcols], weights=dfn['iptw'])
                wn = dwn.mean
            else:
                wt = np.mean(dft[vcols].dropna(), axis=0)
                wn = np.mean(dfn[vcols].dropna(), axis=0)
            return float((wt - wn) / np.sqrt((wt*(1 - wt) + wn*(1 - wn))/2))

        if var_type == 'continuous':
            if weighted:
                dwt = DescrStatsW(dft[vcols], weights=dft['iptw'], ddof=1)
                wmt = dwt.mean
                wst = dwt.std
                dwn = DescrStatsW(dfn[vcols], weights=dfn['iptw'], ddof=1)
                wmn = dwn.mean
                wsn = dwn.std
            else:
                dwt = DescrStatsW(dft[vcols], ddof=1)
                wmt = dwt.mean
                wst = dwt.std
                dwn = DescrStatsW(dfn[vcols], ddof=1)
                wmn = dwn.mean
                wsn = dwn.std
            return float((wmt - wmn) / np.sqrt((wst**2 + wsn**2)/2))

        if var_type == 'categorical':
            if weighted:
                wt = np.average(dft[vcols], weights=dft['iptw'], axis=0)
                wn = np.average(dfn[vcols], weights=dfn['iptw'], axis=0)
            else:
                wt = np.average(dft[vcols], axis=0)
                wn = np.mean(dfn[vcols], axis=0)

            t_c = wt - wn
            s_inv = np.linalg.inv(self._categorical_cov(treated=wt, untreated=wn))
            return float(np.sqrt(np.dot(np.transpose(t_c[1:]), np.dot(s_inv, t_c[1:]))))

    def _weight_calculator(self, df, denominator, numerator):
        """Calculates the IPTW based on the predicted probabilities and the specified group to standardize to in the
        background for the fit() function. Not intended to be used by users

        df is the dataframe, denominator is the string indicating the column of Pr, numerator is the string indicating
        the column of Pr
        """
        if self.stabilized:  # Stabilized weights
            if self.standardize == 'population':
                df['w'] = np.where(df[self.ex] == 1, (df[numerator] / df[denominator]),
                                   ((1 - df[numerator]) / (1 - df[denominator])))
                df['w'] = np.where(df[self.ex].isna(), np.nan, df['w'])
            # Stabilizing to exposed (compares all exposed if they were exposed versus unexposed)
            elif self.standardize == 'exposed':
                df['w'] = np.where(df[self.ex] == 1, 1,
                                   ((df[denominator] / (1 - df[denominator])) * ((1 - df[numerator]) /
                                                                                 df[numerator])))
                df['w'] = np.where(df[self.ex].isna(), np.nan, df['w'])
            # Stabilizing to unexposed (compares all unexposed if they were exposed versus unexposed)
            else:
                df['w'] = np.where(df[self.ex] == 1,
                                   (((1 - df[denominator]) / df[denominator]) * (df[numerator] /
                                                                                 (1 - df[numerator]))),
                                   1)
                df['w'] = np.where(df[self.ex].isna(), np.nan, df['w'])

        else:  # Unstabilized weights
            if self.standardize == 'population':
                df['w'] = np.where(df[self.ex] == 1, 1 / df[denominator], 1 / (1 - df[denominator]))
                df['w'] = np.where(df[self.ex].isna(), np.nan, df['w'])
            # Stabilizing to exposed (compares all exposed if they were exposed versus unexposed)
            elif self.standardize == 'exposed':
                df['w'] = np.where(df[self.ex] == 1, 1, (df[denominator] / (1 - df[denominator])))
                df['w'] = np.where(df[self.ex].isna(), np.nan, df['w'])
            # Stabilizing to unexposed (compares all unexposed if they were exposed versus unexposed)
            else:
                df['w'] = np.where(df[self.ex] == 1, ((1 - df[denominator]) / df[denominator]), 1)
                df['w'] = np.where(df[self.ex].isna(), np.nan, df['w'])
        return df['w']

    @staticmethod
    def _categorical_cov(treated, untreated):
        """Turns out, pandas and numpy don't have the correct covariance matrix I need for categorical variables.
        The covariance matrix is defined as

        S = [S_{kl}] = (P_{1k}*(1-P_{1k}) + P_{2k}*(1-P{2k})) / 2     if k == l
                       (P_{1k}*P_{1l} + P_{2k}*P_{2l}) / 2            if k != l

        Returns covariance matrix
        """
        cv2 = []
        for i, v in enumerate(treated):
            cv1 = []
            if i == 0:
                pass
            else:
                for j, w in enumerate(untreated):
                    if j == 0:
                        pass
                    elif i == j:
                        cv1.append((v * (1 - v) + w * (1 - w)) / 2)
                    else:
                        cv1.append((treated[i] * treated[j] + untreated[i] * untreated[j]) / -2)
                cv2.append(cv1)

        return np.array(cv2)




import numpy as np
import pandas as pd
from scipy import stats
import matplotlib.pyplot as plt


class MonteCarloRR:
    def __init__(self, observed_RR, sd=None, sample=10000):
        """Monte Carlo simulation to assess the impact of an unmeasured binary confounder on the results
        of a study. Observed RR comes from the data analysis, while the RR between the unmeasured confounder
        and the outcome should be obtained from prior literature or constitute an reasonable guess.
        Probability of exposure between the groups should also be reasonable numbers.

        Parameters
        ------------
        observed_RR : float
            Observed RR from the data, not accounting for some binary unmeasured confounder
        sd : float, optional
            Standard deviation of the observed log(risk ratio). This parameter is optional. If specified, then random
            error is incorporated into the bias analysis estimates
        sample : integer, optional
            Number of MC simulations to run. It is important that the specified size of later distributions
            matches this number of samples

        Examples
        -------------
        Monte Carlo bias analysis with trapezoidal distributions
        >>>from zepid.sensitivity_analysis import MonteCarloRR, trapezoidal
        >>>mcrr = MonteCarloRR(observed_RR=0.73322, sample=10000)
        >>>mcrr.confounder_RR_distribution(trapezoidal(mini=0.9, mode1=1.1, mode2=1.7, maxi=1.8, size=10000))
        >>>mcrr.prop_confounder_exposed(trapezoidal(mini=0.25, mode1=0.28, mode2=0.32, maxi=0.35, size=10000))
        >>>mcrr.prop_confounder_unexposed(trapezoidal(mini=0.55, mode1=0.58, mode2=0.62, maxi=0.65, size=10000))
        >>>mcrr.fit()

        Printing a summarization of the bias analysis to the console
        >>>mcrr.summary()

        Creating a density plot of the bias analysis results
        >>>import matplotlib.pyplot as plt
        >>>mcrr.plot()
        >>>plt.show()
        """
        self.RRo = observed_RR
        self._sd = sd
        self.sample = sample

        self.pc0 = None
        self.pc1 = None
        self.RRc_dist = None
        self.corrected_RR = None

    def confounder_RR_distribution(self, dist, seed=None):
        """Distribution of the risk ratio between the unmeasured confounder and the outcome. This
        value should come from prior literature or a reasonable guess. Any numpy random distribution
        can be based to this function. Alternatively, the trapezoid distribution within this library
        can also be used

        Parameters
        ------------
        dist :
            Distribution from which the confounder-outcome Risk Ratio is pulled from. Input should be something like
            `numpy.random.triangular(left=0.9,mode=1.2,right=1.6)` or `zepid.sensitivity_analysis.trapezoidal`
        seed : int, optional
            NumPy seed for the generated distribution. Default is None
        """
        if seed is not None:
            np.random.seed(seed)
        self.RRc_dist = dist

    def prop_confounder_exposed(self, dist, seed=None):
        """Distribution of the proportion of the unmeasured confounder in the exposed group. This value should come
        from prior literature or a reasonable guess. Any numpy random distribution can be based to this function.
        Alternatively, the trapezoid distribution within this library can also be used

        Parameters
        ------------
        dist :
            Distribution from which the confounder-exposure probability is pulled from. Input should be something like
            `numpy.random.triangular(left=0.9,mode=1.2,right=1.6)` or `zepid.sensitivity_analysis.trapezoidal`
        seed : int, optional
            NumPy seed for the generated distribution. Default is None
        """
        if seed is not None:
            np.random.seed(seed)
        self.pc1 = dist

    def prop_confounder_unexposed(self, dist, seed=None):
        """Distribution of the proportion of the unmeasured confounder in the unexposed group. This
        value should come from prior literature or a reasonable guess. Any numpy random distribution
        can be based to this function. Alternatively, the trapezoid distribution within this library
        can also be used

        Parameters
        ------------
        dist :
            Distribution from which the confounder-no exposure probability is pulled from. Input should be something
            like `numpy.random.triangular(left=0.9,mode=1.2,right=1.6)` or `zepid.sensitivity_analysis.trapezoidal`
        seed : int, optional
            NumPy seed for the generated distribution. Default is None
        """
        if seed is not None:
            np.random.seed(seed)
        self.pc0 = dist

    def fit(self):
        """After the observed Risk Ratio, distribution of the confounder-outcome Risk Ratio, proportion
        of the unmeasured confounder in exposed, proportion of the unmeasured confounder in the unexposed.

        .. math::
            RR* = RR / d
            d = (p1 * (RRc-1) + 1) / (p0 * (RRc - 1) + 1)

        Where RR* is the corrected risk ratio, RR is the observed risk ratio in the data set, RRc is the risk ratio
        between unmeasured confounder and outcome, p1 is the probability/proportion of unmeasured confounder in
        exposed, and p0 is the probability/proportion of unmeasured confounder in unexposed
        """
        if self.RRc_dist is None:
            raise ValueError('"confounder_RR_distribution()" has not been specified')
        if self.pc1 is None:
            raise ValueError('"prop_confounder_exposed()" has not been specified')
        if self.pc0 is None:
            raise ValueError('"prop_confounder_unexposed()" has not been specified')
        
        # Monte Carlo
        sf = pd.DataFrame(index=[i for i in range(self.sample)])
        sf['RR_cnf'] = self.RRc_dist
        sf['p1'] = self.pc1
        sf['p0'] = self.pc0
        sf['RR_obs'] = self.RRo
        sf['RR_cor'] = sf['RR_obs'] / ((sf['p1']*(sf['RR_cnf']-1)+1) / (sf['p0'] * (sf['RR_cnf'] - 1) + 1))

        if self._sd is not None:
            sf['RR_cor'] = np.exp(np.log(sf['RR_cor']) - np.random.normal(size=self.sample)*self._sd)

        # Setting new attribute
        self.corrected_RR = np.array(sf['RR_cor'])

    def summary(self, decimal=3):
        """Generate the summary information after the corrected risk ratio distribution is
        generated. fit() must be run before this

        Parameters
        -------------
        decimal : int, optional
            Decimal places to display in output. Default is 3
        """
        print('----------------------------------------------------------------------')
        print('Median corrected Risk Ratio: ', np.round(np.median(self.corrected_RR),decimal))
        print('Mean corrected Risk Ratio: ', np.round(np.mean(self.corrected_RR),decimal))
        print('25th & 75th Percentiles: ', np.round_(np.percentile(self.corrected_RR,q=[25, 75]), decimals=decimal))
        print('2.5th & 97.5th Percentiles: ', np.round_(np.percentile(self.corrected_RR,q=[2.5, 97.5]),
                                                        decimals=decimal))
        print('----------------------------------------------------------------------')

    def plot(self, bw_method='scott', fill=True, color='b'):
        """Generate a Gaussian kernel density plot of the corrected risk ratio distribution. The
        kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or Silverman's Rule can
        be implemented

        Parameters
        -------------
        bw_method : str, optional
            Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
        fill : bool, optional
            Whether to color the area under the density curves. Default is true
        color : str, optional
            Color of the line/area for the treated group. Default is Blue

        Returns
        ------------
        matplotlib axes
        """
        x = np.linspace(np.min(self.corrected_RR),np.max(self.corrected_RR),100)
        density = stats.kde.gaussian_kde(self.corrected_RR,bw_method=bw_method)

        # Creating density plot
        ax = plt.gca()
        if fill:
            ax.fill_between(x, density(x), color=color, alpha=0.2)
        ax.plot(x, density(x), color=color)
        ax.set_xlabel('Corrected Risk Ratio')
        ax.set_ylabel('Density')
        return ax



# utilities for IPW methods. Basically for propensity_score(), which they all use

import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links


def propensity_score(df, model, print_results=True):
    """Generate propensity scores (probability) based on the model input. Uses logistic regression model
    to calculate

    Parameters
    -----------
    df : DataFrame
        Pandas Dataframe containing the variables of interest
    model : str
        Model to fit the logistic regression to. For example, 'y ~ var1 + var2'
    print_results : bool, optional
        Whether to print the logistic regression results. Default is True

    Returns
    -------------
    Fitted statsmodels GLM object

    Example
    ------------
    >>>import zepid as ze
    >>>df = ze.load_sample_data(timevary=False)
    >>>ze.causal.ipw.propensity_score(df=df,model='dead ~ art0 + male + dvl0')
    """
    f = sm.families.family.Binomial()
    log = smf.glm(model, df, family=f).fit()
    if print_results:
        print('\n----------------------------------------------------------------')
        print('MODEL: ' + model)
        print('-----------------------------------------------------------------')
        print(log.summary())
    return log



"""
_____________________________________________________________________________________________
|                                                                                           |
|       zEpid package: making epidemiology with Python e-z                                  |
|                                                                                           |
| This package is to provide easy to use tools to aid in epidemiological analyses. A full   |
| applied guide to the entire package is available at:                                      |
|       https://github.com/pzivich/Python-for-Epidemiologists                               |
|                                                                                           |
| Current contents include:                                                                 |
|       -Basics tools for pandas dataframes and summary data                                |
|       -Useful graphics                                                                    |
|       -Causal estimation methods                                                          |
|       -Sensitivity analysis tools                                                         |
|___________________________________________________________________________________________|

CONTENTS

Basic calculations (both for summary data and pandas.DataFrame):
    -Risk Ratio, Risk Difference, Number Needed to Treat, Odds Ratio, Incidence Rate Ratio, Incidence Rate Difference,
     Interaction Contrast, Interaction Contrast Ratio, Sensitivity, Specificity, Splines

Causal methods:
    -Inverse Probability of Treatment Weights, Inverse Probability of Censoring Weights, Inverse Probability of
     Missing Weights, Time-Fixed G-Formula, Time-Fixed Double Robust
     Estimator

Graphics:
    -Forest Plot, Receiver Operator Curve, Functional Form Plot, P-value Distribution Plot, Spaghetti Plot, Dynamic
     Risk Plot

Sensitivity Analyses:
    -Monte Carlo correction for Risk Ratio, Trapezoidal distribution

See http://zepid.readthedocs.io/en/latest/ for a full guide through all the package features
"""
from .base import (RiskRatio, RiskDifference, NNT, OddsRatio, IncidenceRateRatio, IncidenceRateDifference, Sensitivity,
                   Specificity, Diagnostics, interaction_contrast, interaction_contrast_ratio, spline, table1_generator)
from .datasets import (load_sample_data, load_ewing_sarcoma_data, load_gvhd_data, load_sciatica_data,
                       load_leukemia_data, load_longitudinal_data, load_binge_drinking_data, load_case_control_data,
                       load_monotone_missing_data)

import zepid.calc
import zepid.graphics
import zepid.causal.gformula
import zepid.causal.ipw
import zepid.causal.doublyrobust
import zepid.sensitivity_analysis

from .version import __version__



import warnings
import numpy as np
import pandas as pd

from .utils import propensity_score


class IPCW:
    def __init__(self, df, idvar, time, event, flat_df=False, enter=None):
        """Calculate the inverse probability of censoring weights. Note that this function will accept either a flat
        file (one row per individual) or a long format (multiple rows per individual). If a flat file is provided, it
        must be converted to a long format. This will be done automatically if flat_df is set to True. Additionally, a
        warning and some comparison statistics are provided. Please verify that they match.

        IPC weights are calculated via logistic regression and weights are cumulative products per unique ID. IPCW can
        be used to correct for missing at random data by the generated model in weighted Kaplan-Meier curves. The
        formula used to generate the unstabilized IPCW is

        .. math::

            \pi_i(t) = \cumprod_{R_k \le t} \frac{1}{\Pr(C_i > R_k | \bar{L} = \bar{l}, C_i > R_{k-1})}

        The stabilized IPCW substitutes predicted probabilities under the specified numerator model into the numerator
        of the previous equation. It is similar in concept to IPTW and IPMW in that regards.

        Parameters
        ---------------
        df : DataFrame
            Pandas DataFrame object containing all the variables of interest
        idvar : str
            String that indicates the column name for a unique identifier for each individual
        time : str
            Column name for the ending observation time
        event : str
            Column name for the event of interest
        flat_df : bool, optional
            Whether the input dataframe only contains a single row per participant. If so, the flat dataframe is
            converted to a long dataframe. Default is False (for multiple rows per person)
        enter : str, optional
            Time participant began being observed. Default is None. This option is only needed when flat_df=True

        Example
        ------------
        Setting up the environment
        >>>from zepid import load_sample_data
        >>>from zepid.causal.ipw import IPCW
        >>>df = load_sample_data(timevary=True)
        >>>df['enter_q'] = df['enter'] ** 2
        >>>df['enter_c'] = df['enter'] ** 3
        >>>df['age0_q'] = df['age0'] ** 2
        >>>df['age0_c'] = df['age0'] ** 3

        Calculating IPCW with a long data set
        >>>ipc = IPCW(df, idvar='id', time='enter', event='dead')
        >>>ipc.regression_models(model_denominator='enter + enter_q + enter_c + male + age0 + age0_q + age0_c',
        >>>                      model_numerator='enter + enter_q + enter_c')
        >>>ipc.fit()

        Extracting calculated IPCW
        >>>ipc.Weight

        Calculating IPCW with a wide data set
        >>>df = load_sample_data(False)
        >>>ipc = IPCW(df, idvar='id', time='t', event='dead', flat_df=True)
        >>>ipc.regression_models(model_denominator='enter + enter_q + enter_c + male + age0 + age0_q + age0_c',
        >>>                      model_numerator='enter + enter_q + enter_c')
        >>>ipc.fit()

        """
        if np.sum(df[time].isnull()) > 0:
            raise ValueError('Time is missing for at least one individual in the dataset. They must be removed before '
                             'IPCW can be fit.')
        f = df.copy()
        f.sort_values(by=[idvar, time], inplace=True)
        if np.max(f[time]) == 1:
            raise ValueError('The maximum observation time is 1. For IPCW to function properly, it needs to be greater'
                             'than 1. The more categories, the better the weight estimation')
        if flat_df:
            self.df = self._dataprep(f, idvar, time, event, enter=enter)
        else:
            self.df = f
            self.df['__uncensored__'] = np.where((self.df[idvar] != self.df[idvar].shift(-1)) &
                                                 (self.df[event] == 0),
                                                 0, 1)  # generating indicator for uncensored
            self.df['__uncensored__'] = np.where(self.df[time] == np.max(df[time]), 1, self.df['__uncensored__'])
        self.idvar = idvar
        self.time = time
        self.event = event
        self.Weight = None

    def regression_models(self, model_denominator, model_numerator, print_results=True):
        """Regression model to generate predicted probabilities of censoring, conditional on specified variables.
        Whether stabilized or unstabilized IPCW are generated depends on the specified model numerator.

        Parameters
        --------------
        model_denominator : str
            String of predictor variables for the denominator separated by +. Any variables included in the numerator,
            should be included in the denominator as well. Example 'var1 + var2 + var3 + t_start + t_squared'
        model_numerator : str
            String of predictor variables for the numerator separated by +. In general, time is used as the
            stabilization factor. Example of argument 't_start + t_squared'
        print_results : bool, optional
            Whether to print the model results. Default is True
        """
        nmodel = propensity_score(self.df, '__uncensored__ ~ ' + model_numerator, print_results=print_results)
        self.df['__numer__'] = nmodel.predict(self.df)
        dmodel = propensity_score(self.df, '__uncensored__ ~ ' + model_denominator, print_results=print_results)
        self.df['__denom__'] = dmodel.predict(self.df)
        self.df['__cnumer__'] = self.df.groupby(self.idvar)['__numer__'].cumprod()
        self.df['__cdenom__'] = self.df.groupby(self.idvar)['__denom__'].cumprod()

    def fit(self):
        """Generate the IPC Weights for each observation period for each observation. The calculated weights can be
        accessed through the Weights attribute

        Returns
        -------------
        Fills in the Weight attribute
        """
        self.Weight = self.df['__cnumer__'] / self.df['__cdenom__']

    @staticmethod
    def _dataprep(cf, idvar, time, event, enter=None):
        """Function to prepare the data to an appropriate format for the IPCW class. It breaks the dataset into
        single observations for event one unit increase in time. This is a background process and not meant for users
        to access

        Parameters for my reference
        cf:
            -pandas dataframe to convert into a long format
        idvar:
            -ID variable to retain for observations
        time:
            -last follow-up visit for participant
        event:
            -indicator of whether participant had the event (1 is yes, 0 is no)
        enter:
            -entry time for the participant. Default is None, which means all participants are assumed
             to enter at time zero. Input should be column name of entrance time
        """
        # Copying observations over times
        cf['t_int_zepid'] = cf[time].astype(int)
        lf = pd.DataFrame(np.repeat(cf.values, cf['t_int_zepid'] + 1, axis=0), columns=cf.columns)
        lf['tpoint_zepid'] = lf.groupby(idvar)['t_int_zepid'].cumcount()
        lf['tdiff_zepid'] = lf[time] - lf['tpoint_zepid']
        lf = lf.loc[
            lf['tdiff_zepid'] != 0].copy()  # gets rid of censored at absolute time point (ex. censored at time 10)
        lf.loc[lf['tdiff_zepid'] > 1, 'delta_indicator_zepid'] = 0
        lf.loc[((lf['tdiff_zepid'] <= 1) & (lf[event] == 0)), 'delta_indicator_zepid'] = 0
        lf.loc[((lf['tdiff_zepid'] <= 1) & (lf[event] == 1) & (lf[idvar] != lf[idvar].shift(-1))),
               'delta_indicator_zepid'] = 1
        lf['t_enter_zepid'] = lf['tpoint_zepid']
        lf['t_out_zepid'] = np.where(lf['tdiff_zepid'] < 1, lf[time], lf['t_enter_zepid'] + 1)
        lf['uncensored_zepid'] = np.where((lf[idvar] != lf[idvar].shift(-1)) & (lf['delta_indicator_zepid'] == 0), 0, 1)
        lf['uncensored_zepid'] = np.where(lf['t_out_zepid'] == np.max(lf['t_out_zepid']), 1, lf['uncensored_zepid'])

        # Removing blocks of observations that would have occurred before entrance into the sample
        if enter is not None:
            lf = lf.loc[lf['t_enter_zepid'] >= lf[enter]].copy()

        # Cleaning up the edited dataframe to return to user
        if enter is None:
            lf.drop(['tdiff_zepid', 'tpoint_zepid', 't_int_zepid', time, event], axis=1, inplace=True)
        else:
            lf.drop(['tdiff_zepid', 'tpoint_zepid', 't_int_zepid', time, event, enter], axis=1, inplace=True)
        lf.rename(columns={"delta_indicator_zepid": event, 'uncensored_zepid': '__uncensored__',
                           't_enter_zepid': 't_enter', 't_out_zepid': 't_out'}, inplace=True)
        warnings.warn('Please verify the long dataframe was generated correctly', UserWarning)
        print('Check for dataframe')
        print('\tEvents in input:', np.sum(cf[event]))
        print('\tEvents in output:', np.sum(lf[event]))
        print('\tCensor in input:', cf.dropna(subset=[event]).shape[0] - np.sum(cf[event]))
        print('\tCensor in output:', lf.shape[0] - np.sum(lf['__uncensored__']))
        if enter is None:
            print('\tTotal t input:', np.sum(cf[time]))
            print('\tTotal t output:', np.sum(lf.loc[(lf[idvar] != lf[idvar].shift(-1))]['t_out']))
        else:
            print('\tLate in input:', np.sum(cf[enter] != 0))
            print('\tLate in output:', np.sum(np.where((lf['t_enter'] != 0) & (lf[idvar] != lf[idvar].shift(1)), 1, 0)))
            print('\tTotal t input:', np.sum(cf[time] - cf[enter]))
            print('\tTotal t output:', np.sum(lf['t_out'] - lf['t_enter']))
        return lf.reset_index(drop=True)




import warnings
import numpy as np
import pandas as pd
from .utils import propensity_score


class IPMW:
    def __init__(self, df, missing_variable, stabilized=False, monotone=True):
        """Calculates the weight for inverse probability of missing weights using logistic regression. `IPMW`
        automatically codes a missingness indicator (based on np.nan), so data can be directly input, without creation
        of missingness indicator before inputting data

        `IPMW` currently supports weights for a single missing variable, or a list of variables that are monotonically
        missing. For data to be missing monotonically, there is some ordering of the variables with missing data
        such that the previous variable must be observed for the later to be observed. A simple example is censoring in
        longitudinal data without late entry. To be observed at time *t*, the individual must be observed at time *t-1*

        Nonmonotonic missing data is arguably more common in practice. Sun and Tchetgen Tchetgen recently proposed a
        way to estimate IPMW under nonmonotonic missing data. I plan on implementing this in a future release. Until
        then `IPMW` only supports monotonic missing data

        Parameters
        ----------
        df : DataFrame
            Pandas dataframe object containing all variables of interest
        missing_variable : str, list
            Column name for missing data. numpy.nan values should indicate missing observations. For multiple missing
            variables, a list of strings (indicating column labels) can be added
        stabilized : bool, optional
            Whether to return the stabilized or unstabilized IPMW. Default is to return unstabilized weights
        monotone : bool, optional
            Whether missing data is monotonic or nonmonotonic. This option is only used for when multiple missing
            variables are provided

        Examples
        --------
        Setting up the environment
        >>>from zepid import load_sample_data
        >>>from zepid.causal.ipw import IPMW
        >>>df = load_sample_data(timevary=False)

        Calculating unstabilized Inverse Probability of Missingness Weights
        >>>ipm = IPMW(df, missing='dead', stabilized=False)
        >>>ipm.regression_models(model_denominator='age0 + art + male')
        >>>ipm.fit()

        Extracting calculated weights
        >>>ipm.Weight

        Calculating monotone IPMW
        >>>from zepid import load_monotone_missing_data
        >>>df = load_monotone_missing_data()
        >>>ipm = IPMW(df, missing_variable=['B', 'C'], monotone=True)
        >>>ipm.regression_models(model_denominator=['L + A', 'L + B'])
        >>>ipm.fit()
        >>>ipm.Weight

        Notes
        -----
        For multiple variables with missing data, `IPMW` determines in the two variables are uniform missing. This is a
        special case of monotonic missing data. As a result, `IPMW` can skip over uniform missing data. See the
        following references for further details

        References
        ----------
        Sun, B., Perkins, N. J., Cole, S. R., Harel, O., Mitchell, E. M., Schisterman, E. F., & Tchetgen Tchetgen, E.
        J. (2017). Inverse-probability-weighted estimation for monotone and nonmonotone missing data. American Journal
        of Epidemiology, 187(3), 585-591.

        Perkins, N. J., Cole, S. R., Harel, O., Tchetgen Tchetgen, E. J., Sun, B., Mitchell, E. M., & Schisterman, E.
        F. (2017). Principled approaches to missing data in epidemiologic studies. American Journal of Epidemiology,
        187(3), 568-575.

        Li, L., Shen, C., Li, X., & Robins, J. M. (2013). On weighting approaches for missing data. Statistical Methods
        in Medical Research, 22(1), 14-30.

        Greenland, S., & Finkle, W. D. (1995). A critical look at methods for handling missing covariates in
        epidemiologic regression analyses. American journal of epidemiology, 142(12), 1255-1264.

        Seaman, S. R., & White, I. R. (2013). Review of inverse probability weighting for dealing with missing data.
        Statistical Methods in Medical Research, 22(3), 278-295.
        """
        # Checking input data has missing labeled as np.nan
        if isinstance(missing_variable, str):
            if df.loc[df[missing_variable].isnull(), missing_variable].shape[0] == 0:
                raise ValueError('IPMW requires that missing data is coded as np.nan')
        else:
            for m in missing_variable:
                if df.loc[df[m].isnull(), m].shape[0] == 0:
                    raise ValueError('IPMW requires that missing data is coded as np.nan')

        self.df = df.copy()
        self.missing = missing_variable
        self.stabilized = stabilized
        self.Weight = None
        self._denominator_model = False
        self._monotone = monotone

    def regression_models(self, model_denominator, model_numerator='1', print_results=True):
        """Regression model to generate predicted probabilities of censoring, conditional on specified variables.
        Whether stabilized or unstabilized IPMW are generated depends on the specified model numerator.

        Parameters
        --------------
        model_denominator : str, list
            String of predictor variables for the denominator separated by +. Any variables included in the numerator,
            should be included in the denominator as well. Example 'var1 + var2 + var3 + t_start + t_squared'
        model_numerator : str, optional
            String of predictor variables for the numerator separated by +. In general, time is used as the
            stabilization factor. Example of argument 't_start + t_squared'
        print_results : bool, optional
            Whether to print the model results. Default is True
        """
        if not self.stabilized:
            if model_numerator != '1':
                raise ValueError('Argument for model_numerator is only used for stabilized=True')

        # IPMW for a single missing variable
        if not isinstance(self.missing, list):
            if isinstance(model_denominator, list):
                raise ValueError('For a single missing variable, the model denominator cannot be a list of models')
            self._single_variable(model_denominator, model_numerator=model_numerator, print_results=print_results)

        # IPMW for monotone missing variables
        elif self._monotone:
            # Check if monotone
            self._check_monotone()

            # Check if uniform
            self.missing, overall_uniform = self._check_overall_uniform(df=self.df, miss_vars=self.missing)

            # Uniform monotone missing can safely be treated as only a single variable
            if overall_uniform:
                warnings.warn("It looks like your data is monotone missing data is uniform. Uniform missing data is a "
                              "special case of monotone missing data, where weights need only depend on a single "
                              "variable. The corresponding weights will be generated using the first specified "
                              "regression model")
                self._single_variable(list(model_denominator)[0],
                                      model_numerator=list(model_numerator)[0], print_results=print_results)

            # When not the special case of uniform monotone, use this procedure
            else:
                self._monotone_variables(model_denominator, model_numerator=model_numerator, print_results=print_results)

        # IPMW for nonmonotone missing variables
        else:
            raise ValueError('Non-monotonic IPMW is to be implemented in the future...')

    def fit(self):
        """Provide the regression model to generate the inverse probability of missing weights. The fitted regression
        model will be used to generate the IPW. The weights can be accessed via the IMPW.Weight attribute

        model:
            -statsmodels glm format for modeling data. Independent variables should be predictive of missingness of
             variable of interest. Example) 'var1 + var2 + var3'
        print_results:
            -whether to print the model results. Default is True
        """
        if self._denominator_model is False:
            raise ValueError('No model has been fit to generated predicted probabilities')

        w = self.df['__numer__'] / self.df['__denom__']
        self.Weight = w

    def _check_monotone(self):
        """Background function to check whether data is actually monotone. Raise ValueError if user requests
        monotone=True, but the missing data is nonmonotonic.
        """
        miss_order = list(reversed(self.missing))
        for m in range(len(miss_order)):
            if m == 0 or miss_order[m] == len(miss_order):
                continue

            # Checking adjacent variables
            post = np.where(self.df[miss_order[m]].isnull(), 1, 0)
            prior = np.where(self.df[miss_order[m-1]].isnull(), 1, 0)

            # Post must be zero where prior is zero
            check = np.where((post == 1) & (prior == 0))
            if np.any(check):
                raise ValueError('It looks like your data is non-monotonic. Please check that the missing variables'
                                 'are input in the correct order and that your data is missing monotonically')

    @staticmethod
    def _check_overall_uniform(df, miss_vars):
        """Checks whether all the provided variables with missing data are missing uniformly. If so, then the
        algorithm treats it as a single missing variable. This is a special case of monotone missing data
        """
        # Checks that missing always at same rows in data set
        uniform = pd.Series(np.prod(df[miss_vars].notnull(), axis=1)).equals(df[miss_vars[0]].notnull().astype(int))
        if uniform:
            return miss_vars[0], True
        else:
            return miss_vars, False

    @staticmethod
    def _check_uniform(df, miss1, miss2):
        """Checks whether two variables are missing uniformly. If so, this is a special case of monotone missing data
        that can be treated as a single missing variable. This is used for monotone missing data when all provided
        variables are not missing at random
        """
        # Check that two variables are not uniform missing
        uniform = pd.Series(np.prod(df[[miss1, miss2]].notnull(), axis=1)).equals(df[miss1].notnull().astype(int))
        return uniform

    def _single_variable(self, model_denominator, model_numerator, print_results):
        """Estimates probabilities when only a single variable is missing
        """
        self.df.loc[self.df[self.missing].isnull(), '_observed_indicator_'] = 0
        self.df.loc[self.df[self.missing].notnull(), '_observed_indicator_'] = 1

        dmodel = propensity_score(self.df, '_observed_indicator_ ~ ' + model_denominator, print_results=print_results)
        self.df['__denom__'] = np.where(self.df[self.missing].notnull(), dmodel.predict(self.df), np.nan)
        self._denominator_model = True

        if self.stabilized:
            nmodel = propensity_score(self.df, '_observed_indicator_ ~ ' + model_numerator, print_results=print_results)
            self.df['__numer__'] = np.where(self.df[self.missing].notnull(), nmodel.predict(self.df), np.nan)
        else:
            self.df['__numer__'] = np.where(self.df[self.missing].notnull(), 1, np.nan)

    def _monotone_variables(self, model_denominator, model_numerator, print_results):
        """Estimates probabilities under the monotone missing mechanism
        """
        model_denominator = list(model_denominator)
        model_numerator = list(model_numerator)

        # Check to make sure number of models is not more than number of missing variables
        if len(self.missing) < len(model_denominator) or len(self.missing) < len(model_numerator):
            raise ValueError('More models are specified than missing variables!')

        # If less models than missing variables are specified, repeat the last model for all variables
        while len(self.missing) > len(model_denominator):
            model_denominator.append(model_denominator[-1])
        while len(self.missing) > len(model_numerator):
            model_numerator.append(model_numerator[-1])

        # Looping through all missing variables and specified models
        probs_denom = pd.Series([1] * self.df.shape[0])
        probs_num = pd.Series([1] * self.df.shape[0])

        for mv, model_d, model_n in zip(self.missing, model_denominator, model_numerator):
            df = self.df.copy()

            # Restricting to all those observed by the "outer set" variable
            if mv == self.missing[0]:
                uniform = False
            else:
                # Checking to see if this variable and the previous are uniformly missing
                uniform = self._check_uniform(df, miss1=self.missing[self.missing.index(mv) - 1], miss2=mv)
                # Restricting to only observed
                df = df.loc[df[self.missing[self.missing.index(mv) - 1]].notnull()].copy()

            if uniform:
                continue
            else:
                df.loc[df[mv].isnull(), '_observed_indicator_'] = 0
                df.loc[df[mv].notnull(), '_observed_indicator_'] = 1
                dmodel = propensity_score(df, '_observed_indicator_ ~ ' + model_d, print_results=print_results)
                probs_denom = probs_denom * dmodel.predict(self.df)

                # Only for stabilized IPMW with monotone missing data
                if self.stabilized:
                    nmodel = propensity_score(df, '_observed_indicator_ ~ ' + model_n, print_results=print_results)
                    probs_num = probs_num * nmodel.predict(self.df)

        # Calculating Probabilities
        self.df['__denom__'] = np.where(self.df[self.missing[-1]].notnull(), probs_denom, np.nan)
        if self.stabilized:
            self.df['__numer__'] = np.where(self.df[self.missing[-1]].notnull(), probs_num, np.nan)
        else:
            self.df['__numer__'] = np.where(self.df[self.missing[-1]].notnull(), 1, np.nan)
        self._denominator_model = True

